/*
 * Copyright (c) 2021-2022 Nvidia, Inc. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of Ford forlkenson max flow algorithm.
 * This file implements the algorithm. Currently only for trees with
 * the same speeds on all of the links.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_MAX_FLOW_ALGORITHM_C
#include <opensm/osm_max_flow_algorithm.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>

static inline boolean_t flow_is_residual(flow_edge_t *p_edge)
{
	return (p_edge->capacity == 0);
}

static inline int flow_get_remaining_capacity(flow_edge_t *p_edge)
{
	return (p_edge->capacity - p_edge->flow);
}


static inline void flow_set_capacity(flow_edge_t *p_edge, int capacity)
{
	p_edge->capacity = capacity;
	p_edge->residual_edge->capacity = 0;
}

static inline void flow_augment(flow_edge_t *p_edge, int bottle_neck)
{
	p_edge->flow += bottle_neck;
	p_edge->residual_edge->flow -= bottle_neck;
}

static flow_edge_t *flow_add_new_edge(flow_vertex_t *vertex)
{
	vertex->degree++;
	return &vertex->edges[vertex->degree - 1];
}

static void flow_init_edge(flow_edge_t *p_edge, int index, int port_num,
			   int adj_index, int adj_port_num, int capacity)
{
	p_edge->index = index;
	p_edge->port_num = port_num;
	p_edge->adj_index = adj_index;
	p_edge->adj_port_num = adj_port_num;
	p_edge->capacity = capacity;
	p_edge->flow = 0;
}

static void flow_add_edge(flow_vertex_t *vertices, int index, int port_num,
			  int adj_index, int adj_port_num, int capacity)
{
	flow_edge_t *p_edge = flow_add_new_edge(&vertices[index]);
	flow_edge_t *p_residual_edge = flow_add_new_edge(&vertices[adj_index]);

	flow_init_edge(p_edge, index, port_num,
		       adj_index, adj_port_num, capacity);
	flow_init_edge(p_residual_edge, adj_index, adj_port_num,
		       index, port_num, capacity);

	p_edge->residual_edge = p_residual_edge;
	p_residual_edge->residual_edge = p_edge;
}

static void destroy_count_sort_array(flow_graph_t *p_graph)
{
	int i;
	int array_size = p_graph->count_sort_array_size;
	cl_list_t *count_sort = p_graph->count_sort;

	if (!count_sort)
		return;

	for (i = 0; i < array_size; i++) {
		cl_list_t *p_list = &count_sort[i];

		cl_list_remove_all(p_list);
		cl_list_destroy(p_list);
	}

	free(p_graph->count_sort);
	p_graph->count_sort = NULL;
}

void osm_flow_destroy_graph(flow_graph_t **pp_graph)
{
	flow_graph_t *p_graph = *pp_graph;
	flow_vertex_t *vertices;
	void **prev_privs;
	osm_switch_t *p_sw;
	int i;

	if (!p_graph)
		return;

	if (p_graph->prev_privs) {
		vertices = p_graph->vertices;
		prev_privs = p_graph->prev_privs;
		for (i = 0; i < p_graph->vertices_num; i++) {
			p_sw = vertices[i].p_sw;
			p_sw->priv = prev_privs[i];
		}

		free(p_graph->prev_privs);
	}

	free(p_graph->visited);
	free(p_graph->dirty);
	free(p_graph->vertices);
	destroy_count_sort_array(p_graph);
	free(p_graph);
	*pp_graph = NULL;
}

static flow_graph_t *flow_allocate_graph(osm_subn_t *p_subn,
					 boolean_t save_prev_privs)
{
	flow_graph_t *p_graph = NULL;
	flow_vertex_t *vertices = NULL;
	int *visited = NULL;
	boolean_t *dirty = NULL;
	void **prev_privs = NULL;
	unsigned int num_vertices = 0;
	int i;

	num_vertices = cl_qmap_count(&p_subn->sw_guid_tbl);
	if (num_vertices == 0) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9735: Cannot route topology without switches\n");
		goto Exit;
	}

	p_graph = (flow_graph_t *)calloc(1, sizeof(flow_graph_t));
	if (!p_graph) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9736: Cannot allocate memory for graph\n");
		goto Exit;
	}

	/* Create vertices array. */
	vertices = (flow_vertex_t *)malloc(num_vertices * sizeof(flow_vertex_t));
	if (!vertices) {
		free(p_graph);
		p_graph = NULL;
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9737: Cannot allocate memory for vertices\n");
		goto Exit;
	}

	for (i = 0; i < num_vertices; i++) {
		flow_vertex_t *p_vertex = &vertices[i];

		p_vertex->rank = 0;
		p_vertex->degree = 0;
		p_vertex->calculate_flow = FALSE;
	}

	visited = (int *)calloc(num_vertices, sizeof(int));
	if (!visited) {
		free(vertices);
		free(p_graph);
		p_graph = NULL;
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9738: Cannot allocate memory for visited array\n");
		goto Exit;
	}

	dirty = (int *)calloc(num_vertices, sizeof(int));
	if (!dirty) {
		free(visited);
		free(vertices);
		free(p_graph);
		p_graph = NULL;
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9739: Cannot allocate memory for dirty array\n");
		goto Exit;
	}

	if (save_prev_privs) {
		prev_privs = (void **)calloc(num_vertices, sizeof(void *));
		if (!prev_privs) {
			free(dirty);
			free(visited);
			free(vertices);
			free(p_graph);
			p_graph = NULL;
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR,
				"ERR 9740: Cannot allocate memory for prev_privs array\n");
			goto Exit;
		}
		p_graph->prev_privs = prev_privs;
	}

	p_graph->vertices = vertices;
	p_graph->visited = visited;
	p_graph->dirty = dirty;
	p_graph->vertices_num = num_vertices;
	p_graph->p_subn = p_subn;
	p_graph->count_sort = NULL;

Exit:
	return p_graph;
}

flow_graph_t *osm_flow_build_graph(osm_subn_t *p_subn)
{
	cl_map_item_t *item;
	osm_switch_t *p_sw, *p_rem_sw;
	osm_node_t *p_node, *p_rem_node;
	flow_vertex_t *vertices = NULL;
	flow_vertex_t *p_vertex, *p_adj_vertex;
	unsigned int num_vertices;
	unsigned int index;
	uint8_t port_num, rem_port_num;
	flow_graph_t *p_graph = NULL;
	int *visited;
	osm_physp_t *p_physp;
	void **prev_privs;
	int flow_calculations_num = 0;

	p_graph = flow_allocate_graph(p_subn, TRUE);
	if (!p_graph)
		goto Exit;

	num_vertices = p_graph->vertices_num;
	visited = p_graph->visited;
	vertices = p_graph->vertices;
	prev_privs = p_graph->prev_privs;

	index = 0;
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		p_vertex = &vertices[index];
		p_vertex->p_sw = p_sw;
		prev_privs[index] = p_sw->priv;
		p_sw->priv = p_vertex;
		p_vertex->index = index;
		index++;
	}

	for (index = 0; index < num_vertices; index++) {
		boolean_t hca_exists = FALSE;
		boolean_t is_leaf = TRUE;

		p_vertex = &vertices[index];
		visited[index] = 1;
		p_sw = p_vertex->p_sw;
		p_node = p_sw->p_node;
		for (port_num = 1; port_num < p_sw->num_ports; port_num++) {
			int adj_index;

			p_rem_node = osm_node_get_remote_node(p_node, port_num,
							      &rem_port_num);
			if (!p_rem_node || p_rem_node == p_node)
				continue;

			if (osm_node_get_type(p_rem_node)  == IB_NODE_TYPE_ROUTER)
				continue;

			if (!p_rem_node->sw) {
				p_physp = osm_node_get_physp_ptr(p_node, port_num);
				if (!osm_phsyp_get_special_port_type(p_physp->p_remote_physp))
					hca_exists = TRUE;

				continue;
			}

			p_rem_sw = p_rem_node->sw;
			if (p_sw->rank < p_rem_sw->rank)
				is_leaf = FALSE;

			p_adj_vertex = (flow_vertex_t *)p_rem_sw->priv;
			adj_index = p_adj_vertex->index;
			if (!visited[adj_index])
				flow_add_edge(vertices, index, port_num,
					      adj_index, rem_port_num, 1);
		}

		if (is_leaf && hca_exists) {
			flow_calculations_num++;
			p_vertex->calculate_flow = TRUE;
		}
	}

	/*
	 * To know what is the maximum number of times a port can be part of an
	 * augmented path in one max flow calculation need to calculate:
	 * max_augment_port_picks = max(c1,c2,...) / gcd(c1,c2,...)
	 * c1,c2... stands for the capacity/speed of the links.
	 * Need to multiply the flow_calculations_num by max_augment_port_picks
	 * to know the maximum number a port can be picked in all
	 * the calculations from a specific source,
	 * Currently, the algorithm supports only topologies with the same
	 * speed on all the links, so the value of max_augment_port_picks is 1.
	 */
	p_graph->count_sort_array_size = flow_calculations_num;

Exit:
	return p_graph;
}

static cl_status_t flow_init_count_sort_array(flow_graph_t *p_graph)
{
	int i;
	cl_status_t status = CL_SUCCESS;

	p_graph->count_sort = (cl_list_t *)calloc(p_graph->count_sort_array_size,
						  sizeof(cl_list_t));
	if (!p_graph->count_sort) {
		OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 3339: Cannot allocate memory for count_sort array\n");
		status = CL_INSUFFICIENT_RESOURCES;
		goto Exit;
	}

	for (i = 0; i < p_graph->count_sort_array_size; i++)
		cl_list_init(&p_graph->count_sort[i], 10);

Exit:
	return status;
}

flow_graph_t *osm_flow_copy_graph(flow_graph_t *p_graph)
{
	flow_graph_t *p_new_graph = flow_allocate_graph(p_graph->p_subn, FALSE);
	int num_vertices, index, edge_index;
	flow_vertex_t *new_vertices, *vertices;
	flow_vertex_t *p_new_vertex, *p_vertex;
	flow_vertex_t **new_sw_lid_to_vertex_mapping;
	flow_edge_t *edges;
	flow_edge_t *p_edge;
	int *new_visited;
	u_int16_t sw_lid;

	if (!p_new_graph)
		goto Exit;

	p_new_graph->count_sort_array_size = p_graph->count_sort_array_size;
	vertices = p_graph->vertices;
	num_vertices = p_new_graph->vertices_num;
	new_vertices = p_new_graph->vertices;
	new_visited = p_new_graph->visited;
	new_sw_lid_to_vertex_mapping = p_new_graph->sw_lid_to_vertex_mapping;

	for (index = 0; index < num_vertices; index++) {
		p_new_vertex = &new_vertices[index];
		p_vertex = &vertices[index];
		new_visited[index] = 1;

		p_new_vertex->index = index;
		p_new_vertex->calculate_flow = p_vertex->calculate_flow;
		p_new_vertex->p_sw = p_vertex->p_sw;
		sw_lid = cl_ntoh16(osm_node_get_base_lid(p_new_vertex->p_sw->p_node, 0));
		new_sw_lid_to_vertex_mapping[sw_lid] = p_new_vertex;

		edges = p_vertex->edges;
		for (edge_index = 0; edge_index < p_vertex->degree; edge_index++) {
			p_edge = &edges[edge_index];

			if (!new_visited[p_edge->adj_index])
				flow_add_edge(new_vertices, p_edge->index, p_edge->port_num,
					      p_edge->adj_index, p_edge->adj_port_num, 1);
		}
	}

	if (flow_init_count_sort_array(p_new_graph) != CL_SUCCESS)
		osm_flow_destroy_graph(&p_new_graph);

Exit:
	return p_new_graph;
}

static int _flow_dfs_solve(flow_graph_t *p_graph, flow_vertex_t *p_vertex,
			   int flow, boolean_t is_src_vertex, int *start_idx)
{
	flow_vertex_t *vertices = p_graph->vertices;
	uint16_t dst_lid_ho = p_graph->dst_lid_ho;
	int *visited = p_graph->visited;
	boolean_t *dirty = p_graph->dirty;
	int min_flow, bottle_neck, adj_index;
	int remaining_capacity;
	unsigned int i;
	flow_edge_t *p_edge = NULL;
	flow_vertex_t *p_adj_vertex;
	flow_edge_t *edges;
	uint16_t min_hops_vertex, hops_adj_vertex;
	uint8_t **hops;
	cl_list_iterator_t next_list_iter = NULL;
	cl_list_iterator_t cur_list_iter = NULL;
	cl_list_t *p_list = NULL;
	cl_list_t *p_tmp_list = NULL;
	int j = *start_idx;
	int count_sort_array_size = p_graph->count_sort_array_size;

	if (p_vertex == p_graph->p_dst_vertex)
		return flow;

	visited[p_vertex->index] = p_graph->visited_token;
	min_hops_vertex = osm_switch_get_least_hops(p_vertex->p_sw, dst_lid_ho);
	hops = p_vertex->p_sw->hops;

	edges = p_vertex->edges;
	if (is_src_vertex) {
		p_list = &p_graph->count_sort[j];
		next_list_iter = cl_list_head(p_list);
		cur_list_iter = next_list_iter;
	}

	for (i = 0; i < p_vertex->degree; i++) {
		p_edge = NULL;
		if (is_src_vertex) {
			while (j < count_sort_array_size) {
				if (next_list_iter != cl_list_end(p_list)) {
					p_edge = (flow_edge_t *) cl_list_obj(next_list_iter);
					cur_list_iter = next_list_iter;
					next_list_iter = cl_list_next(cur_list_iter);
					break;
				} else {
					if ((j == *start_idx) && !cl_list_count(p_list))
						*start_idx = j + 1;

					j++;
					if (j < count_sort_array_size) {
						p_list = &p_graph->count_sort[j];
						next_list_iter = cl_list_head(p_list);
					}
				}
			}

			if (!p_edge) {
				OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
					"ERR 3340: edge missing in count sort array\n");
				return 0;
			}
		} else {
			p_edge = &edges[i];
		}

		p_adj_vertex = &vertices[p_edge->adj_index];
		adj_index = p_adj_vertex->index;

		remaining_capacity = flow_get_remaining_capacity(p_edge);
		hops_adj_vertex = hops[dst_lid_ho][p_edge->port_num];
		if (remaining_capacity && (visited[adj_index] != p_graph->visited_token) &&
		    ((min_hops_vertex == hops_adj_vertex) || flow_is_residual(p_edge))) {
			min_flow = remaining_capacity > flow ? flow : remaining_capacity;
			bottle_neck = _flow_dfs_solve(p_graph, p_adj_vertex, min_flow, FALSE, start_idx);

			if (bottle_neck > 0) {
				flow_augment(p_edge, bottle_neck);
				dirty[adj_index] = TRUE;
				if (is_src_vertex) {
					cl_list_remove_item(p_list, cur_list_iter);
					p_tmp_list = &p_graph->count_sort[j + bottle_neck];
					cl_list_insert_tail(p_tmp_list, p_edge);
				}

				return bottle_neck;
			}
		}
	}

	return 0;
}


static int flow_dfs_solve(flow_graph_t *p_graph, flow_vertex_t *p_src_vertex, int *start_idx)
{
	if (p_src_vertex == p_graph->p_dst_vertex)
		return 0;

	return _flow_dfs_solve(p_graph, p_src_vertex, INT32_MAX, TRUE, start_idx);
}

void flow_edge_print(flow_graph_t *p_graph, flow_edge_t *p_edge)
{
	flow_vertex_t *vertices = p_graph->vertices;
	flow_vertex_t *p_vertex = &vertices[p_edge->index];
	flow_vertex_t *p_adj_vertex = &vertices[p_edge->adj_index];

	OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
		"Edge 0x%016" PRIx64 "[%d] --> 0x%016" PRIx64
		"[%d], flow = %d, capacity = %d, is_residual = %d\n",
		cl_ntoh64(osm_node_get_node_guid(p_vertex->p_sw->p_node)), p_edge->port_num,
		cl_ntoh64(osm_node_get_node_guid(p_adj_vertex->p_sw->p_node)),
		p_edge->adj_port_num, p_edge->flow, p_edge->capacity, flow_is_residual(p_edge));
}

void flow_print_edges(flow_graph_t *p_graph)
{
	flow_vertex_t *vertices = p_graph->vertices;
	flow_vertex_t *p_vertex;
	int index, e;

	for (index = 0; index < p_graph->vertices_num; index++) {
		p_vertex = &vertices[index];

		for (e = 0; e < p_vertex->degree; e++)
			flow_edge_print(p_graph, &p_vertex->edges[e]);
	}
}

int osm_flow_max_flow_solve(flow_graph_t *p_graph, flow_vertex_t *p_src_vertex,
			    flow_vertex_t *p_dst_vertex, int *start_idx)
{
	int f;
	int max_flow = 0;
	uint16_t dst_lid_ho;
	uint16_t hops_vertex;

	p_graph->visited_token = 2;
	p_graph->p_dst_vertex = p_dst_vertex;

	dst_lid_ho = cl_ntoh16(osm_node_get_base_lid(p_graph->p_dst_vertex->p_sw->p_node, 0));
	hops_vertex = osm_switch_get_least_hops(p_src_vertex->p_sw, dst_lid_ho);
	if (hops_vertex == OSM_NO_PATH) {
		OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_ERROR,
			"ERR 9742: src SW: 0x%016" PRIx64
			", dst SW: 0x%016" PRIx64 ", max_flow: %d\n",
			cl_ntoh64(osm_node_get_node_guid(p_src_vertex->p_sw->p_node)),
			cl_ntoh64(osm_node_get_node_guid(p_dst_vertex->p_sw->p_node)), 0);

		return 0;
	}

	p_graph->dst_lid_ho = dst_lid_ho;

	for (f = flow_dfs_solve(p_graph, p_src_vertex, start_idx);
	     f != 0; f = flow_dfs_solve(p_graph, p_src_vertex, start_idx)) {
		p_graph->visited_token++;
		max_flow += f;
	}

	OSM_LOG(&p_graph->p_subn->p_osm->log, OSM_LOG_DEBUG, "Src SW: "
		"0x%016" PRIx64 ", Dst SW: 0x%016" PRIx64
		", visited_token: %d, max_flow: %d\n",
		cl_ntoh64(osm_node_get_node_guid(p_src_vertex->p_sw->p_node)),
		cl_ntoh64(osm_node_get_node_guid(p_dst_vertex->p_sw->p_node)),
		p_graph->visited_token, max_flow);

	return max_flow;
}



int osm_flow_update_edge_directions(flow_graph_t *p_graph, flow_vertex_t *p_src_vertex)
{
	int result = 0;
	unsigned int queue_in = 0;
	unsigned int queue_out = 0;
	osm_log_t *p_log = &p_graph->p_subn->p_osm->log;
	flow_vertex_t **queue = NULL;
	flow_vertex_t *vertices = p_graph->vertices;
	int *visited = p_graph->visited;
	flow_vertex_t *p_vertex, *p_adj_vertex;
	flow_edge_t *p_edge;
	int visited_token;
	int edge_index, index, adj_index;

	queue = (flow_vertex_t **)calloc(p_graph->vertices_num, sizeof(flow_vertex_t*));
	if (!queue) {
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR 9741: Cannot allocate memory for routing"
			" engine\n");
		result = 1;
		goto Exit;
	}

	p_graph->visited_token = 1;
	visited_token = p_graph->visited_token;
	memset(visited, 0, sizeof(int) * p_graph->vertices_num);
	for (index = 0; index < p_graph->vertices_num; index++)
		vertices[index].rank = 0;

	queue[queue_in++] = p_src_vertex;
	visited[p_src_vertex->index] = p_graph->visited_token;

	while (queue_out < queue_in) {
		p_vertex = queue[queue_out++];

		/* Add adjacent vertices to queue. */
		for (edge_index = 0; edge_index < p_vertex->degree; edge_index++) {
			p_edge = &p_vertex->edges[edge_index];
			p_adj_vertex = &vertices[p_edge->adj_index];
			adj_index = p_adj_vertex->index;
			if (visited[adj_index] == visited_token) {
				if (p_adj_vertex->rank > p_vertex->rank)
					flow_set_capacity(p_edge, 1);
				else
					flow_set_capacity(p_edge->residual_edge, 1);

				continue;
			}

			flow_set_capacity(p_edge, 1);
			visited[adj_index] = visited_token;
			p_adj_vertex->rank = p_vertex->rank + 1;
			queue[queue_in++] = p_adj_vertex;
		}
	}

	free(queue);
Exit:
	return result;
}

void osm_flow_init_flow_and_visited(flow_graph_t *p_graph)
{
	flow_vertex_t *vertices = p_graph->vertices;
	boolean_t *dirty = p_graph->dirty;
	flow_vertex_t *p_vertex;
	flow_edge_t *p_edge;
	int index, e;

	memset(p_graph->visited, 0, sizeof(int) * p_graph->vertices_num);
	for (index = 0; index < p_graph->vertices_num; index++) {
		if (!dirty[index])
			continue;

		p_vertex = &vertices[index];
		dirty[index] = FALSE;
		for (e = 0; e < p_vertex->degree; e++) {
			p_edge = &p_vertex->edges[e];
			p_edge->flow = 0;
			p_edge->residual_edge->flow = 0;
		}
	}
}

