/*
 * Copyright (c) 2004-2008 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2013 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 * Copyright (c) 2009 HNR Consulting. All rights reserved.
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of Up Down Algorithm using ranking & Min Hop
 *      Calculation functions
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <ctype.h>
#include <complib/cl_debug.h>
#include <complib/cl_qmap.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UCAST_UPDN_C
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_mcast_mgr.h>
#include <opensm/osm_root_detection.h>

extern int osm_apply_func_on_port_group_nodes(char *pgrp_name, int pfn_func(void *, uint64_t , char *),
                                              void *cxt, char *pi, void *p_osm, boolean_t goto_node);

/* //////////////////////////// */
/*  Local types                 */
/* //////////////////////////// */

/* direction */
typedef enum updn_switch_dir {
	UP = 0,
	DOWN
} updn_switch_dir_t;


typedef struct updn_minhop_data {
	cl_list_item_t list;
	struct osm_switch *sw;
	uint64_t id;
	updn_switch_dir_t dir;
	uint8_t visited;
	void *prev_priv;
} updn_minhop_data_t;

typedef uint64_t color_t;

#define NO_COLOR 0
#define MAX_LEAF_COLOR 0x8000000000000000
#define MAX_COLOR UINT64_MAX


typedef struct updn_node {
	cl_list_item_t list;
	osm_switch_t *sw;
	uint8_t num_hca;
	boolean_t is_leaf;
	uint32_t *dlid_routes;
	uint32_t *dlid_sw_routes;
	uint32_t total_routes;
	color_t color;
} updn_node_t;

typedef struct endport {
	osm_port_t *p_port;
	updn_node_t *adj_sw;
	uint8_t path_bit;
} endport_t;

typedef struct group_offset {
	int offset;
	color_t color;
} group_offset_t;


/* updn structure */
typedef struct updn {
	unsigned num_roots;
	osm_opensm_t *p_osm;
	osm_subn_t *p_subn;
	unsigned num_nodes;
	updn_node_t **ranked_nodes;
	group_offset_t *offset_arr;
	int offset_arr_len;
	endport_t *port_arr;
	int port_arr_len;
} updn_t;

static inline char *direction_str(updn_switch_dir_t direction)
{
	return direction == UP ? "UP" : "DOWN";
}

static inline osm_subn_t *updn_get_subnet(IN updn_t * p_updn)
{
	return p_updn->p_subn;
}

/* This function returns direction based on rank and guid info of current &
   remote ports */
static updn_switch_dir_t updn_get_dir(unsigned cur_rank, unsigned rem_rank,
				      uint64_t cur_id, uint64_t rem_id,
				      uint8_t hops)
{
	/* HACK: comes to solve root nodes connection, in a classic subnet root nodes do not connect
	   directly, but in case they are we assign to root node an UP direction to allow UPDN to discover
	   the subnet correctly (and not from the point of view of the last root node).
	 */
	if (!cur_rank && !rem_rank)
		return UP;

	if (cur_rank < rem_rank) {
		/* allow one hop between "siblings" on a way down */
		if (hops > (rem_rank - cur_rank + 1))
			return UP;
		return DOWN;
	} else if (cur_rank > rem_rank) {
		return UP;
	} else {
		/* Equal rank, decide by id number, bigger == UP direction */
		if (cur_id > rem_id || hops > 1)
			return UP;
		else
			return DOWN;
	}
}

/**********************************************************************
 * This function does the bfs of min hop table calculation by guid index
 * as a starting point.
 **********************************************************************/
static int updn_bfs_by_node(IN osm_log_t * p_log, IN osm_subn_t * p_subn,
			    IN osm_switch_t * p_sw, uint32_t id)
{
	uint8_t pn, pn_rem;
	cl_qlist_t list;
	uint16_t lid;
	updn_switch_dir_t next_dir, current_dir;
	updn_minhop_data_t *u;
	OSM_LOG_ENTER(p_log);

	lid = osm_node_get_base_lid(p_sw->p_node, 0);
	lid = cl_ntoh16(lid);
	osm_switch_set_hops(p_sw, lid, 0, 0);

	OSM_LOG(p_log, OSM_LOG_DEBUG,
		"Starting from switch - port GUID 0x%" PRIx64 " lid %u\n",
		cl_ntoh64(p_sw->p_node->node_info.port_guid), lid);
	u = (updn_minhop_data_t *) (p_sw->priv);
	u = &u[id];
	u->dir = UP;

	/* Update list with the new element */
	cl_qlist_init(&list);
	cl_qlist_insert_tail(&list, &u->list);

	/* BFS the list till no next element */
	while (!cl_is_qlist_empty(&list)) {
		u = (updn_minhop_data_t *)cl_qlist_remove_head(&list);
		u->visited = 0;	/* cleanup */
		current_dir = u->dir;
		/* Go over all ports of the switch and find unvisited remote nodes */
		for (pn = 1; pn < u->sw->num_ports; pn++) {
			osm_node_t *p_remote_node;
			updn_minhop_data_t *rem_u;
			uint8_t current_min_hop, remote_min_hop,
			    set_hop_return_value;
			osm_switch_t *p_remote_sw;
			osm_physp_t *p_physp;
			p_remote_node =
			    osm_node_get_remote_node(u->sw->p_node, pn, &pn_rem);

			p_physp = osm_node_get_physp_ptr(u->sw->p_node, pn);

			/* If no remote node OR remote node is not a SWITCH
			   continue to next pn */
			if (!p_remote_node || !p_remote_node->sw ||
			    !p_physp || !osm_link_is_healthy(p_physp))
				continue;

			/* Fetch remote guid only after validation of remote node */
			p_remote_sw = p_remote_node->sw;
			rem_u = (updn_minhop_data_t *) p_remote_sw->priv;
			rem_u = &rem_u[id];
			/* Decide which direction to mark it (UP/DOWN) */
			next_dir = updn_get_dir(u->sw->rank, rem_u->sw->rank,
						u->sw->id, rem_u->sw->id, 1);

			/* Check if this is a legal step : the only illegal step is going
			   from DOWN to UP */
			if ((current_dir == DOWN) && (next_dir == UP)) {
				OSM_LOG(p_log, OSM_LOG_DEBUG,
					"Avoiding move from 0x%016" PRIx64
					" to 0x%016" PRIx64 "\n",
					cl_ntoh64(osm_node_get_node_guid(u->sw->p_node)),
					cl_ntoh64(osm_node_get_node_guid(p_remote_node)));
				/* Illegal step */
				continue;
			}
			/* Set MinHop value for the current lid */
			current_min_hop = osm_switch_get_least_hops(u->sw, lid);
			/* Check hop count if better insert into list && update
			   the remote node Min Hop Table */
			remote_min_hop =
			    osm_switch_get_hop_count(p_remote_sw, lid, pn_rem);
			if (current_min_hop + 1 < remote_min_hop) {
				set_hop_return_value =
				    osm_switch_set_hops(p_remote_sw, lid,
							pn_rem,
							current_min_hop + 1);
				if (set_hop_return_value) {
					OSM_LOG(p_log, OSM_LOG_ERROR, "ERR AA01: "
						"Invalid value returned from set min hop is: %d\n",
						set_hop_return_value);
				}
				/* Check if remote port has already been visited */
				if (!rem_u->visited) {
					/* Insert updn_switch item into the list */
					rem_u->dir = next_dir;
					rem_u->visited = 1;
					cl_qlist_insert_tail(&list,
							     &rem_u->list);
				}
			}
		}
	}

	OSM_LOG_EXIT(p_log);
	return 0;
}

static void updn_bfs_by_node_par(IN void *context, uint32_t my_id)
{
	osm_ucast_mgr_t *p_mgr = context;
	cl_list_item_t * item;
	osm_switch_t * p_sw;
	cl_qlist_t * p_list;

	p_list = &p_mgr->sw_lists[my_id];

	while (cl_qlist_count(p_list)) {

		item = cl_qlist_remove_head(p_list);

		p_sw = cl_item_obj(item, p_sw, list_item);

		updn_bfs_by_node(p_mgr->p_log, p_mgr->p_subn, p_sw, my_id);
	}
}

/* NOTE : PLS check if we need to decide that the first */
/*        rank is a SWITCH for BFS purpose */
static int updn_subn_rank(IN updn_t * p_updn)
{
	osm_switch_t *p_sw;
	osm_physp_t *p_physp, *p_remote_physp;
	cl_qlist_t list;
	cl_map_item_t *item;
	struct updn_node *u, *remote_u;
	uint8_t num_ports, port_num;
	osm_log_t *p_log = &p_updn->p_osm->log;
	unsigned max_rank = 0;
	osm_subn_t *p_subn;
	boolean_t is_leaf;

	OSM_LOG_ENTER(p_log);
	cl_qlist_init(&list);

	p_subn = updn_get_subnet(p_updn);

	/* add all roots to the list */
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		u = p_sw->priv;
		if (!p_sw->rank)
			cl_qlist_insert_tail(&list, &u->list);
	}

	/* BFS the list till it's empty */
	while (!cl_is_qlist_empty(&list)) {
		u = (struct updn_node *)cl_qlist_remove_head(&list);
		is_leaf = TRUE;
		/* Go over all remote nodes and rank them (if not already visited) */
		p_sw = u->sw;
		num_ports = p_sw->num_ports;
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"Handling switch GUID 0x%" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
		for (port_num = 1; port_num < num_ports; port_num++) {
			ib_net64_t port_guid;

			/* Current port fetched in order to get remote side */
			p_physp =
			    osm_node_get_physp_ptr(p_sw->p_node, port_num);

			if (!p_physp)
				continue;

			p_remote_physp = p_physp->p_remote_physp;

			/*
			   make sure that all the following occur on p_remote_physp:
			   1. The port isn't NULL
			   2. It is a switch
			 */
			if (!p_remote_physp)
				continue;

			if (p_remote_physp->p_node->sw) {
				remote_u = p_remote_physp->p_node->sw->priv;
				port_guid = p_remote_physp->port_guid;

				if (remote_u->sw->rank > u->sw->rank)
					is_leaf = FALSE;

				if (remote_u->sw->rank > u->sw->rank + 1) {
					/* Rank assigned to switch only once */
					OSM_ASSERT(remote_u->sw->rank ==
						   OSM_SW_NO_RANK);
					remote_u->sw->rank = u->sw->rank + 1;
					max_rank = remote_u->sw->rank;
					cl_qlist_insert_tail(&list,
							     &remote_u->list);
					OSM_LOG(p_log, OSM_LOG_DEBUG,
						"Rank of port GUID 0x%" PRIx64
						" = %u\n", cl_ntoh64(port_guid),
						remote_u->sw->rank);
				}
				continue;
			}

			if (osm_node_get_type(p_remote_physp->p_node) != IB_NODE_TYPE_CA)
				continue;

			if (osm_phsyp_get_special_port_type(p_remote_physp))
				continue;

			u->num_hca++;
		}

		if (is_leaf) {
			osm_subn_add_leaf(p_subn, p_sw);
			u->is_leaf = TRUE;
		}
	}

	/* Print Summary of ranking */
	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Subnet ranking completed. Max Node Rank = %d\n", max_rank);
	OSM_LOG_EXIT(p_log);
	return 0;
}


/* hack: preserve min hops entries to any other root switches */
static void updn_clear_non_root_hops(IN osm_subn_t * p_subn, osm_switch_t * sw)
{
	osm_port_t *port;
	unsigned i;

	for (i = 0; i < sw->num_hops; i++)
		if (sw->hops[i]) {
			port = osm_get_port_by_lid_ho(p_subn, i);
			if (!port || !port->p_node->sw ||
			    port->p_node->sw->rank)
				memset(sw->hops[i], 0xff, sw->num_ports);
		}
}

int updn_calculate_min_hop_table(IN osm_subn_t *p_subn)
{
	osm_log_t *p_log = &p_subn->p_osm->log;
	osm_switch_t *p_sw;
	cl_map_item_t *item;
	uint32_t i, number_of_threads = p_subn->opt.routing_threads_num;
	uint32_t sw_idx, successfull_alloc_switch = 0;
	int res = 0;
	cl_pc_thread_pool_t pc_thread_pool;
	osm_ucast_mgr_t * p_mgr = &p_subn->p_osm->sm.ucast_mgr;
	updn_minhop_data_t *u = NULL;

	OSM_LOG_ENTER(p_log);

	/* Go over all the switches in the subnet - for each init their Min Hop
	   Table */
	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Init Min Hop Table of all switches [\n");

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		/* Clear Min Hop Table */
		if (p_subn->opt.connect_roots)
			updn_clear_non_root_hops(p_subn, p_sw);
		else
			osm_switch_clear_hops(p_sw);
	}

	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Init Min Hop Table of all switches ]\n");

	/* Now do the BFS for each port  in the subnet */
	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"BFS through all port guids in the subnet [\n");

	if (number_of_threads != 1) {

		res = osm_ucast_mgr_pc_thread_pool_init(p_mgr,
							&pc_thread_pool,
							&updn_bfs_by_node_par,
							"minhop_calculation",
							&number_of_threads);
		if (res != 0)
			goto Exit;
	}

	/* create sw->priv array per thread */
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl), successfull_alloc_switch = 0;
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item), successfull_alloc_switch++) {

		p_sw = (osm_switch_t *)item;
		u = (updn_minhop_data_t *) calloc (1, sizeof(updn_minhop_data_t) * number_of_threads);
		if (u == NULL) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR AA11: "
				"Not enough memory for priv objects\n");
			res = -1;
			goto Exit;
		}
		for (i = 0; i < number_of_threads; i++) {
			u[i].sw = p_sw;
			u[i].prev_priv = p_sw->priv;
		}
		/* change the pointer */
		p_sw->priv = u;
	}

	if (number_of_threads == 1) {
		for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
		     item != cl_qmap_end(&p_subn->sw_guid_tbl);
		     item = cl_qmap_next(item)) {
			p_sw = (osm_switch_t *)item;
			updn_bfs_by_node(p_log, p_subn, p_sw, 0);
		}
		res = 0;
		goto Exit;
	}

	/* Multithreaded part of the algorithm */
	cl_pc_thread_pool_start(&pc_thread_pool);

	cl_pc_thread_pool_wait(&pc_thread_pool, EVENT_NO_TIMEOUT);
Exit:
	/* restore previous value for switches that changed */
	for (sw_idx = 0, item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     sw_idx < successfull_alloc_switch && item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     sw_idx++, item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		u = (updn_minhop_data_t *) p_sw->priv;
		/* restore previous data */
		p_sw->priv = u[0].prev_priv;
		/* free previously allocated data for minhop calculation */
		free(u);
	}

	if (number_of_threads != 1) {
		cl_pc_thread_pool_destroy(&pc_thread_pool);
		free(p_mgr->sw_lists);
		p_mgr->sw_lists = NULL;
	}
	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"BFS through all port guids in the subnet ]\n");
	/* Cleanup */
	OSM_LOG_EXIT(p_log);
	return res;
}

static int updn_build_lid_matrices(IN updn_t * p_updn)
{
	int status;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);

	OSM_LOG_ENTER(&p_updn->p_osm->log);

	OSM_LOG(&p_updn->p_osm->log, OSM_LOG_VERBOSE,
		"Ranking all port guids in the list\n");
	if (!p_updn->num_roots) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA0A: "
			"No guids were provided or number of guids is 0\n");
		status = -1;
		goto _exit;
	}

	/* Check if it's not a switched subnet */
	if (cl_is_qmap_empty(&p_subn->sw_guid_tbl)) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA0B: "
			"This is not a switched subnet, cannot perform UPDN algorithm\n");
		status = -1;
		goto _exit;
	}

	/* Rank the subnet switches */
	if (updn_subn_rank(p_updn)) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA0E: "
			"Failed to assign ranks\n");
		status = -1;
		goto _exit;
	}

	/* After multiple ranking need to set Min Hop Table by UpDn algorithm  */
	OSM_LOG(&p_updn->p_osm->log, OSM_LOG_VERBOSE,
		"Setting all switches' Min Hop Table\n");
	status = updn_calculate_min_hop_table(p_subn);

_exit:
	OSM_LOG_EXIT(&p_updn->p_osm->log);
	return status;
}

static struct updn_node *create_updn_node(osm_switch_t * sw)
{
	struct updn_node *u;

	u = malloc(sizeof(*u));
	if (!u)
		return NULL;
	memset(u, 0, sizeof(*u));
	u->sw = sw;
	u->is_leaf = FALSE;
	u->num_hca = 0;
	u->dlid_routes = NULL;
	u->dlid_sw_routes = NULL;
	u->total_routes = 0;
	sw->rank = OSM_SW_NO_RANK;
	return u;
}

static void delete_updn_node(struct updn_node *u)
{
	u->sw->priv = NULL;
	free(u->dlid_routes);
	free(u->dlid_sw_routes);
	free(u);
}

static void updn_delete_all(updn_t *p_updn)
{
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	cl_map_item_t *item;
	osm_switch_t *p_sw;

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *) item;
		delete_updn_node(p_sw->priv);
	}

	free(p_updn->ranked_nodes);
	p_updn->ranked_nodes = NULL;

	free(p_updn->port_arr);
	free(p_updn->offset_arr);
	p_updn->port_arr = NULL;
	p_updn->offset_arr = NULL;
	p_updn->port_arr_len = 0;
	p_updn->offset_arr_len = 0;
}

/* Find Root nodes automatically by Min Hop Table info */
static void updn_find_root_nodes_by_min_hop(OUT updn_t * p_updn)
{
	osm_opensm_t *p_osm = p_updn->p_osm;
	osm_switch_t *p_sw;
	osm_port_t *p_port;
	osm_physp_t *p_physp;
	cl_map_item_t *item;
	double thd1, thd2;
	unsigned i, cas_num = 0;
	unsigned *cas_per_sw;
	uint16_t lid_ho;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);

	OSM_LOG_ENTER(&p_osm->log);

	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"Current number of ports in the subnet is %d\n",
		cl_qmap_count(&p_subn->port_guid_tbl));

	lid_ho = p_subn->max_lid + 1;
	cas_per_sw = malloc(lid_ho * sizeof(*cas_per_sw));
	if (!cas_per_sw) {
		OSM_LOG(&p_osm->log, OSM_LOG_ERROR, "ERR AA14: "
			"cannot alloc mem for CAs per switch counter array\n");
		goto _exit;
	}
	memset(cas_per_sw, 0, lid_ho * sizeof(*cas_per_sw));

	/* Find the Maximum number of CAs (and routers) for histogram normalization */
	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"Finding the number of CAs and storing them in cl_map\n");
	for (item = cl_qmap_head(&p_subn->port_guid_tbl);
	     item != cl_qmap_end(&p_subn->port_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_port = (osm_port_t *)item;
		if (!p_port->p_node->sw) {
			p_physp = p_port->p_physp->p_remote_physp;
			if (!p_physp || !p_physp->p_node->sw ||
			    osm_phsyp_get_special_port_type(p_port->p_physp))
				continue;
			lid_ho = osm_node_get_base_lid(p_physp->p_node, 0);
			lid_ho = cl_ntoh16(lid_ho);
			cas_per_sw[lid_ho]++;
			cas_num++;
		}
	}

	thd1 = cas_num * 0.9;
	thd2 = cas_num * 0.05;
	OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
		"Found %u CAs and RTRs, %u SWs in the subnet. "
		"Thresholds are thd1 = %f && thd2 = %f\n",
		cas_num, cl_qmap_count(&p_subn->sw_guid_tbl), thd1, thd2);

	OSM_LOG(&p_osm->log, OSM_LOG_VERBOSE,
		"Passing through all switches to collect Min Hop info\n");
	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		unsigned hop_hist[IB_SUBNET_PATH_HOPS_MAX];
		uint16_t max_lid_ho;
		uint8_t hop_val;
		uint16_t numHopBarsOverThd1 = 0;
		uint16_t numHopBarsOverThd2 = 0;

		p_sw = (osm_switch_t *) item;

		memset(hop_hist, 0, sizeof(hop_hist));

		max_lid_ho = p_sw->max_lid_ho;
		for (lid_ho = 1; lid_ho <= max_lid_ho; lid_ho++)
			if (cas_per_sw[lid_ho]) {
				hop_val =
				    osm_switch_get_least_hops(p_sw, lid_ho);
				if (hop_val >= IB_SUBNET_PATH_HOPS_MAX)
					continue;

				hop_hist[hop_val] += cas_per_sw[lid_ho];
			}

		/* Now recognize the spines by requiring one bar to be
		   above 90% of the number of CAs and RTRs */
		for (i = 0; i < IB_SUBNET_PATH_HOPS_MAX; i++) {
			if (hop_hist[i] > thd1)
				numHopBarsOverThd1++;
			if (hop_hist[i] > thd2)
				numHopBarsOverThd2++;
		}

		/* If thd conditions are valid - rank the root node */
		if (numHopBarsOverThd1 == 1 && numHopBarsOverThd2 == 1) {
			OSM_LOG(&p_osm->log, OSM_LOG_DEBUG,
				"Ranking GUID 0x%" PRIx64 " as root node\n",
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			p_sw->rank = 0;
			p_updn->num_roots++;
		}
	}

	free(cas_per_sw);
_exit:
	OSM_LOG_EXIT(&p_osm->log);
	return;
}

static void dump_roots(cl_map_item_t *item, FILE *file, void *cxt)
{
	osm_switch_t *sw = (osm_switch_t *)item;
	if (!sw->rank)
		fprintf(file, "0x%" PRIx64 "\n",
			cl_ntoh64(osm_node_get_node_guid(sw->p_node)));
}

static int rank_root_node(void *cxt, uint64_t guid, char *p)
{
	updn_t *updn = cxt;
	osm_switch_t *sw;

	sw = osm_get_switch_by_guid(updn_get_subnet(updn),
				    cl_hton64(guid));
	if (!sw) {
		OSM_LOG(&updn->p_osm->log, OSM_LOG_ERROR, "ERR AA04: "
			"Root switch with guid 0x%" PRIx64 " is not found\n", guid);
		return 0;
	}

	OSM_LOG(&updn->p_osm->log, OSM_LOG_DEBUG,
		"Ranking root port GUID 0x%" PRIx64 "\n", guid);

	sw->rank = 0;
	updn->num_roots++;

	return 0;
}

/* UPDN callback function */
static int updn_lid_matrices(void *ctx)
{
	updn_t *p_updn = ctx;
	cl_map_item_t *item;
	osm_switch_t *p_sw;
	int ret = 0;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	char root_file_path[1024];

	OSM_LOG_ENTER(&p_updn->p_osm->log);

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		p_sw->priv = create_updn_node(p_sw);
		if (!p_sw->priv) {
			OSM_LOG(&(p_updn->p_osm->log), OSM_LOG_ERROR, "ERR AA0C: "
				"cannot create updn node\n");
			OSM_LOG_EXIT(&p_updn->p_osm->log);
			return -1;
		}
	}

	/* First setup root nodes */
	p_updn->num_roots = 0;

	if (p_subn->opt.root_guid_file) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_DEBUG,
			"UPDN - Fetching root nodes from file \'%s\'\n",
			p_subn->opt.root_guid_file);

		ret = parse_node_map_str(p_subn->opt.root_guid_file,
					 osm_apply_func_on_port_group_nodes, rank_root_node,
					 p_updn, p_updn->p_osm, TRUE);
		if (ret) {
			OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR, "ERR AA02: "
				"cannot parse root guids file \'%s\'\n",
				p_subn->opt.root_guid_file);
			osm_ucast_mgr_build_lid_matrices(&p_updn->p_osm->sm.ucast_mgr);

			if (p_subn->opt.find_roots_color_algorithm)
				osm_root_detection_find_root_nodes_by_coloring(
				    p_updn->p_osm,
				    updn_get_subnet(p_updn),
				    OSM_TOPO_TREE,
				    &p_updn->num_roots);
			else
				updn_find_root_nodes_by_min_hop(p_updn);

		} else if (p_subn->opt.connect_roots &&
			   p_updn->num_roots > 1)
			osm_ucast_mgr_build_lid_matrices(&p_updn->p_osm->sm.ucast_mgr);
	} else {
		osm_ucast_mgr_build_lid_matrices(&p_updn->p_osm->sm.ucast_mgr);

		if (p_subn->opt.find_roots_color_algorithm)
			osm_root_detection_find_root_nodes_by_coloring(
			    p_updn->p_osm,
			    updn_get_subnet(p_updn),
			    OSM_TOPO_TREE,
			    &p_updn->num_roots);
		else
			updn_find_root_nodes_by_min_hop(p_updn);
	}

	/* Only if there are assigned root nodes do the algorithm, otherwise perform do nothing */
	if (p_updn->num_roots) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_DEBUG,
			"activating UPDN algorithm\n");
		ret = updn_build_lid_matrices(p_updn);
	} else {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_INFO,
			"disabling UPDN algorithm, no root nodes were found\n");
		ret = -1;
	}

	if (OSM_LOG_IS_ACTIVE_V2(&p_updn->p_osm->log, OSM_LOG_ROUTING)) {
		snprintf(root_file_path, sizeof(root_file_path), "%s/opensm-updn-roots.dump",
			 p_updn->p_osm->subn.opt.dump_files_dir);
		osm_dump_qmap_to_file(p_updn->p_osm, root_file_path,
				      &p_subn->sw_guid_tbl,
				      dump_roots, NULL);
	}

	if (ret)
		updn_delete_all(p_updn);

	OSM_LOG_EXIT(&p_updn->p_osm->log);
	return ret;
}
/*
 * Go through sorted ranked switches for DOWN to TOP and skip leafs
 * 1. For each non-leaf, check whether switches below it already have colors.
 *    Create color for the given switch as bitwise OR of colors below it
 * 2. Assign combined color to the leafs attached to the given switch
 * 3. If no leafs are colored yet, assign new color to the given switch
 *    and apply #2
*/
static void updn_assign_color(updn_t *p_updn)
{
	int i, pn;
	color_t tmp_color = NO_COLOR;
	color_t next_color = 1;
	updn_node_t *u;
	osm_log_t *p_log = &p_updn->p_osm->log;

	OSM_LOG_ENTER(p_log);

	for (i = p_updn->num_nodes - 1; i >= 0; i--) {
		u = p_updn->ranked_nodes[i];
		/* skip leafs */
		if (u->is_leaf && u->sw->rank)
			continue;

		tmp_color = 0;
		/* find downwards switches with assigned color */
		for (pn = 1; pn < u->sw->num_ports; pn++) {
			osm_node_t *p_remote_node;
			struct updn_node *rem_u;
			osm_switch_t *p_remote_sw;
			osm_physp_t *p_physp;
			p_remote_node =
			    osm_node_get_remote_node(u->sw->p_node, pn,
						     NULL);

			p_physp = osm_node_get_physp_ptr(u->sw->p_node, pn);

			/* If no remote node OR remote node is not a SWITCH
			   continue to next pn */
			if (!p_remote_node || !p_remote_node->sw ||
			    !p_physp || !osm_link_is_healthy(p_physp))
				continue;

			p_remote_sw = p_remote_node->sw;
			rem_u = (struct updn_node *) p_remote_sw->priv;
			rem_u = &rem_u[0];

			if (DOWN != updn_get_dir(u->sw->rank, rem_u->sw->rank,
						u->sw->id, rem_u->sw->id, 1))
				continue;
			tmp_color |= rem_u->color;
		}
		/* no colors assigned to the children */
		if (tmp_color == NO_COLOR) {
			if (next_color == MAX_LEAF_COLOR) {
				OSM_LOG(p_log, OSM_LOG_ERROR,
					"ERR AA21: More than 64 colors required."
					 "Set zero color \n");
				next_color = NO_COLOR;
			} else {
				tmp_color = next_color;
				next_color <<= 1;
			}
		}
		u->color = tmp_color;

		/* assign color to the children w/o color */
		for (pn = 1; pn < u->sw->num_ports; pn++) {
			osm_node_t *p_remote_node;
			struct updn_node *rem_u;
			osm_switch_t *p_remote_sw;
			osm_physp_t *p_physp;
			p_remote_node =
			    osm_node_get_remote_node(u->sw->p_node, pn,
						     NULL);

			p_physp = osm_node_get_physp_ptr(u->sw->p_node, pn);

			/* If no remote node OR remote node is not a SWITCH
			   continue to next pn */
			if (!p_remote_node || !p_remote_node->sw ||
			    !p_physp || !osm_link_is_healthy(p_physp))
				continue;

			p_remote_sw = p_remote_node->sw;
			rem_u = (struct updn_node *) p_remote_sw->priv;
			rem_u = &rem_u[0];

			if (DOWN != updn_get_dir(u->sw->rank, rem_u->sw->rank,
						u->sw->id, rem_u->sw->id, 1))
				continue;
			if (rem_u->color == NO_COLOR)
				rem_u->color = u->color;
		}
	}
	OSM_LOG_EXIT(p_log);
}

static int create_port_array(updn_t *updn, cl_qlist_t *port_list,
			     uint8_t min_path_bit,
			     uint8_t max_path_bit)
{
	int i = 0, j = 0, port_array_len;
	cl_list_item_t *list_item;
	endport_t *port_array = NULL;
	osm_log_t *p_log = &updn->p_osm->log;
	OSM_LOG_ENTER(p_log);

	/* create endpoint array */
	if (!port_list)
		return -1;

	port_array_len = IB_LID_UCAST_END_HO + 1;

	port_array = malloc(port_array_len * sizeof(endport_t));
	if (!port_array)
		return -1;

	updn->offset_arr = calloc(1,
			   port_array_len * sizeof(group_offset_t));
	updn->offset_arr_len = 0;

	if (!updn->offset_arr) {
		free(port_array);
		return -1;
	}

	for (list_item = cl_qlist_head(port_list);
	     list_item != cl_qlist_end(port_list);
	     list_item = cl_qlist_next(list_item)) {
		osm_port_t *p_dest_port =
			cl_item_obj(list_item, p_dest_port, list_item);
		osm_switch_t *dest_sw;
		osm_physp_t *p_dest_sw_physp, *p_dest_physp;
		uint8_t end_lid_offset;

		if (!p_dest_port->p_node->sw) {
			p_dest_sw_physp =
				osm_physp_get_remote(p_dest_port->p_physp);
			if (!p_dest_sw_physp)
				continue;
			dest_sw = p_dest_sw_physp->p_node->sw;
		}
		else
			dest_sw = p_dest_port->p_node->sw;

		p_dest_physp = osm_port_get_physp(p_dest_port);

		end_lid_offset = (1 << osm_physp_get_lmc(p_dest_physp)) - 1;
		if (min_path_bit > end_lid_offset)
			continue;

		if (max_path_bit < end_lid_offset)
			end_lid_offset = max_path_bit;

		for(j = min_path_bit; j <= end_lid_offset; j++) {
			port_array[i].path_bit = j;
			port_array[i].p_port = p_dest_port;
			port_array[i++].adj_sw = (updn_node_t *) dest_sw->priv;
		}
	}
	updn->port_arr_len = i;
	updn->port_arr = port_array;
	OSM_LOG_EXIT(p_log);
	return 0;

}

/*
 * When converge_routes is TRUE, an adjacent switch with higher value of dlid_routes
 * should be selected. Otherwise an adjacent switch with lower value of dlid_routes
 * should be chosen
*/

inline static boolean_t updn_lid_tracking_compare_dlid_routes(
					 uint32_t curr_dlid_routes,
					 uint32_t adj_sw_dlid_routes,
					 boolean_t converge_routes)
{
	return (converge_routes) ?
		(curr_dlid_routes < adj_sw_dlid_routes):
		(curr_dlid_routes > adj_sw_dlid_routes);
}

static int updn_lid_tracking_converge_routes(updn_t *p_updn, struct updn_node *u)
{
	osm_ucast_mgr_t *p_mgr = &p_updn->p_osm->sm.ucast_mgr;
	osm_switch_t *p_sw = u->sw;
	osm_port_t *p_dest_port;
	osm_physp_t *p_exit_physp;
	osm_node_t *p_adj_node, *curr_adj_node;
	osm_switch_t *p_adj_sw, *p_dest_sw;
	updn_node_t *adj_node;
	boolean_t lft_set;
	uint16_t dest_port_base_lid, dest_port_lid;
	int p, port, exit_port, i;
	uint32_t weights[IB_NODE_NUM_PORTS_MAX] = { 0 };
	uint8_t path_bit;

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Processing routes converge for switch 0x%" PRIx64 "\n",
		cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));

	for (i = 0; i < p_updn->port_arr_len; i++) {

		p_dest_port = p_updn->port_arr[i].p_port;
		path_bit = p_updn->port_arr[i].path_bit;

		dest_port_base_lid = cl_ntoh16(p_dest_port->lid);
		if (!dest_port_base_lid)
			continue;

		dest_port_lid = dest_port_base_lid + path_bit;

		if (p_dest_port->p_node->sw)
			continue;

		/* non-leaf switch - some routes pass to destination */
		if (u->dlid_routes[dest_port_lid])
			continue;

		if (p_sw->new_lft[dest_port_lid] == OSM_NO_PATH)
			continue;

		p_dest_sw = (p_updn->port_arr[i].adj_sw)->sw;

		if (!p_dest_sw)
			continue;

		/* Destination port connected directly to this switch */
		if (p_dest_sw == p_sw)
			continue;

		lft_set = FALSE;
		exit_port = OSM_NO_PATH;
		curr_adj_node = NULL;

		for (p = 1; !lft_set && p < p_sw->num_ports; p++) {
			port = osm_switch_get_dimn_port(p_sw, p);
			if (!port)
				continue;
			/* Check if port is connected and healthy */
			p_exit_physp = osm_node_get_physp_ptr(p_sw->p_node, port);
			if (!p_exit_physp || !osm_physp_is_healthy(p_exit_physp) ||
			    !p_exit_physp->p_remote_physp)
				continue;

			/* Get remote switch */
			p_adj_node = osm_physp_get_node_ptr(p_exit_physp->p_remote_physp);

			p_adj_sw = p_adj_node->sw;

			if (!p_adj_sw)
				continue;

			adj_node = (updn_node_t *)p_adj_sw->priv;

			/* Validate port direction */
			if (UP !=
			    updn_get_dir(p_sw->rank, p_adj_sw->rank,
					 u->sw->id, adj_node->sw->id, 1))
				continue;
			/* No real path through adj_node to destination */
			if (adj_node->dlid_routes[dest_port_lid] == 0)
				continue;
			/* if more than one switch are delivering DLID */
			/* stick to the first one found */
			if (curr_adj_node && curr_adj_node != p_adj_node)
				continue;
			/*
			* take care of parallel links to the
			* switch processing the node
			*/
			if (p_sw->new_lft[dest_port_lid] == port)
				lft_set = TRUE;
			else {
				if ((exit_port == OSM_NO_PATH) ||
				    weights[port] < weights[exit_port]) {
					exit_port = port;
					curr_adj_node = p_adj_node;
				}
			}
		}
		/* converge route */
		if(!lft_set && exit_port != OSM_NO_PATH) {
			p_sw->new_lft[dest_port_lid] = exit_port;
			weights[exit_port]++;
		}

	}

	return 0;
}
static int updn_lid_tracking_process_switch(updn_t *p_updn, struct updn_node *u,
					    enum updn_switch_dir direction,
					    int port_array_start)
{
	osm_ucast_mgr_t *p_mgr = &p_updn->p_osm->sm.ucast_mgr;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	endport_t *port_arr = p_updn->port_arr;
	int port_arr_len = p_updn->port_arr_len;
	osm_switch_t *p_sw = u->sw;
	osm_port_t *p_dest_port;
	osm_physp_t *p_dest_physp, *p_dest_sw_physp = NULL, *p_exit_physp;
	osm_node_t *p_adj_node;
	osm_switch_t *p_dest_sw, *p_adj_sw;
	updn_node_t *exit_node, *adj_node;
	uint32_t weights[IB_NODE_NUM_PORTS_MAX] = { 0 };
	uint32_t weights_in_color[IB_NODE_NUM_PORTS_MAX] = { 0 };
	uint32_t dlid_routes;
	uint32_t dlid_sw_routes;
	uint32_t total_routes, num_routes;
	uint16_t dest_sw_lid, dest_port_base_lid, dest_port_lid;
	uint8_t least_hops, port, p, num_ports = p_sw->num_ports;
	uint8_t exit_port;
	int i;
	color_t prev_color = NO_COLOR;
	uint8_t path_bit;
	boolean_t found, port_on_direction_exist;

	OSM_ASSERT(p_sw->rank != OSM_SW_NO_RANK);

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Processing switch 0x%" PRIx64 " direction %s offset %d\n",
		cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
		direction_str(direction),
		port_array_start);

	for (i = port_array_start; i < port_arr_len + port_array_start; i++) {

		exit_port = OSM_NO_PATH;
		exit_node = NULL;
		dlid_sw_routes = 0;
		dlid_routes = 0;
		total_routes = 0;
		if (prev_color != port_arr[i % port_arr_len].adj_sw->color) {
			prev_color = port_arr[i % port_arr_len].adj_sw->color;
			memset(weights_in_color, 0, sizeof (weights_in_color));
		}
		p_dest_port = port_arr[i % port_arr_len].p_port;
		path_bit = port_arr[i % port_arr_len].path_bit;

		p_dest_physp = osm_port_get_physp(p_dest_port);
		dest_port_base_lid = cl_ntoh16(p_dest_port->lid);
		if (!dest_port_base_lid)
			continue;

		dest_port_lid = dest_port_base_lid + path_bit;

		/* Destination is this switch */
		if (p_dest_port->p_node == p_sw->p_node) {
			if (direction == UP) {
				p_sw->new_lft[dest_port_lid] = 0;
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"Routing LID %u to port 0 for switch"
					" 0x%" PRIx64 "\n",
					dest_port_lid,
					cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			}
			continue;
		}
		p_dest_sw = port_arr[i % port_arr_len].adj_sw->sw;

		/* Destination port connected directly to this switch */
		if (p_dest_sw == p_sw) {
			if (direction == UP) {
				p_dest_sw_physp = osm_physp_get_remote(p_dest_physp);
				OSM_ASSERT(p_dest_sw_physp);
				p_sw->new_lft[dest_port_lid] =
				    osm_physp_get_port_num(p_dest_sw_physp);
			}
			continue;
		}

		if (p_dest_sw->rank == OSM_SW_NO_RANK)
			/* Unranked switch ... */
			continue;

		dest_sw_lid = cl_ntoh16(osm_node_get_base_lid(p_dest_sw->p_node, 0));
		least_hops = osm_switch_get_least_hops(p_sw, dest_sw_lid);
		if (least_hops == OSM_NO_PATH) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"No path to get to LID %u from switch"
				" 0x%" PRIx64 "\n",
				dest_port_lid,
				cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			continue;
		}

		port_on_direction_exist = FALSE;

		for (p = 1; p < num_ports + 1; p++) {
			port = osm_switch_get_dimn_port(p_sw, p % num_ports);
			if (!port)
				continue;

			/* Check if port leads to shortest route */
			if (osm_switch_get_hop_count(p_sw, dest_sw_lid, port) !=
						     least_hops)
				continue;

			/* Check if port is connected and healthy */
			p_exit_physp = osm_node_get_physp_ptr(p_sw->p_node, port);
			if (!p_exit_physp || !osm_physp_is_healthy(p_exit_physp) ||
			    !p_exit_physp->p_remote_physp)
				continue;

			/* Get remote switch */
			p_adj_node = osm_node_get_remote_node(p_sw->p_node,
							      port, NULL);
			OSM_ASSERT(p_adj_node && p_adj_node->sw);
			p_adj_sw = p_adj_node->sw;
			adj_node = (updn_node_t *)p_adj_sw->priv;

			/* Validate port direction.
			 * Don't skip the port in case the link
			 * is between "siblings".
			 */
			if (p_sw->rank != p_adj_sw->rank &&
			    direction !=
			    updn_get_dir(p_sw->rank, p_adj_sw->rank,
					 u->sw->id, adj_node->sw->id, 1))
				continue;
			/* found port on the same direction */
			port_on_direction_exist = TRUE;

			/*
			 * Exit port selection criterias by precedence:
			 * 1. Exit port that leads to switch with maximal/minimal
			 *    number of routes to destination port.
			 *    (depends on opt.updn_lid_tracking_converge_routes)
			 * 2. Exit port with least routes to the nodes
			 *    with the same color as the current node
			 * 3. Exit port with least routes.
			 * 4. Exit port that leads to the switch with
			 *    lesser number of routes passing through it
			 * 5. Exit port that leads to a switch with least routes
			 *    that also go through the switch connected to the
			 *    destination port.
			 */
			if (p_subn->opt.updn_lid_tracking_prefer_total_routes)
				found = (exit_port == OSM_NO_PATH ||
				updn_lid_tracking_compare_dlid_routes(
					dlid_routes,
					adj_node->dlid_routes[dest_port_lid],
					p_subn->opt.updn_lid_tracking_converge_routes) ||
				(dlid_routes == adj_node->dlid_routes[dest_port_lid] &&
				(weights_in_color[exit_port] > weights_in_color[port] ||
				(weights_in_color[exit_port] == weights_in_color[port] &&
				(weights[exit_port] > weights[port] ||
				(weights[exit_port] == weights[port] &&
				(total_routes > adj_node->total_routes ||
				(total_routes == adj_node->total_routes &&
				dlid_sw_routes > adj_node->dlid_sw_routes[dest_sw_lid]))))))));
			else
				found = (exit_port == OSM_NO_PATH ||
				updn_lid_tracking_compare_dlid_routes(
					dlid_routes,
					adj_node->dlid_routes[dest_port_lid],
					p_subn->opt.updn_lid_tracking_converge_routes) ||
				(dlid_routes == adj_node->dlid_routes[dest_port_lid] &&
				(weights_in_color[exit_port] > weights_in_color[port] ||
				(weights_in_color[exit_port] == weights_in_color[port] &&
				(weights[exit_port] > weights[port] ||
				(weights[exit_port] == weights[port] &&
				(dlid_sw_routes > adj_node->dlid_sw_routes[dest_sw_lid] ||
				(dlid_sw_routes == adj_node->dlid_sw_routes[dest_sw_lid] &&
				total_routes > adj_node->total_routes))))))));

			if (found) {
				exit_port = port;
				exit_node = adj_node;
				total_routes = adj_node->total_routes;
				dlid_sw_routes = adj_node->dlid_sw_routes[dest_sw_lid];
				dlid_routes = adj_node->dlid_routes[dest_port_lid];
			}
		}

		if (exit_port == OSM_NO_PATH) {
			/*
			 * only if we had at least one exit port to adjacent switch
			 * on the same direction as the explored direction,
			 * and no valid exit port was found report error
			 */
			if (port_on_direction_exist)
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR AA20: Failed to find %s path to get to LID %u"
					" from switch 0x%" PRIx64 "\n",
					direction_str(direction),
					dest_port_lid,
					cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)));
			continue;
		}

		p_sw->new_lft[dest_port_lid] = exit_port;
		/* Update weights */
		num_routes = u->dlid_routes[dest_port_lid] +
			     (u->num_hca ? u->num_hca : !!u->is_leaf);

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Routing %s LID %u via port %u for switch 0x%" PRIx64 " NUM_R %d to switch %s\n",
			direction_str(direction),
			dest_port_lid, exit_port,
			cl_ntoh64(osm_node_get_node_guid(p_sw->p_node)),
			num_routes,
			exit_node->sw->p_node->print_desc);


		if (num_routes && !p_dest_port->p_node->sw) {
			exit_node->dlid_routes[dest_port_lid] += num_routes;
			exit_node->dlid_sw_routes[dest_sw_lid] += num_routes;
			exit_node->total_routes += num_routes;
			weights[exit_port] += num_routes;
			weights_in_color[exit_port] += num_routes;
		}
	}

	return 0;
}
/*
 * Array is already sorted by nodes under the same switch
 * The switch port0 nodes are at the end of the array.
 * We need to group all nodes with same color together
 * and keep switches at the end
 */

static int updn_port_color_cmp_func(const void *lhs, const void* rhs)
{
	endport_t *p_left = (endport_t *)lhs;
	endport_t *p_right = (endport_t *)rhs;

	if (p_left->p_port->p_node->sw && p_right->p_port->p_node->sw)
		return 0;

	if (p_left->p_port->p_node->sw && !p_right->p_port->p_node->sw)
		return 1;

	if (!p_left->p_port->p_node->sw && p_right->p_port->p_node->sw)
		return -1;

	if (p_left->adj_sw->color < p_right->adj_sw->color)
		return -1;

	if (p_left->adj_sw->color > p_right->adj_sw->color)
		return 1;
	return 0;
}

static int updn_node_rank_cmp_func(const void *lhs, const void* rhs)
{
	updn_node_t *p_left = *(updn_node_t **)lhs;
	updn_node_t *p_right = *(updn_node_t **)rhs;

	return (updn_get_dir(p_left->sw->rank, p_right->sw->rank,
			     p_left->sw->id, p_right->sw->id, 1) == DOWN ? -1 : 1);
}

static int updn_node_load_and_color_cmp_up_func(const void *lhs, const void* rhs)
{
	updn_node_t *p_left = *(updn_node_t **)lhs;
	updn_node_t *p_right = *(updn_node_t **)rhs;
	unsigned lrank = p_left->sw->rank;
	unsigned rrank = p_right->sw->rank;
	color_t lcolor = p_left->color;
	color_t rcolor = p_right->color;
	uint64_t l_id = p_left->sw->id;
	uint64_t r_id = p_right->sw->id;
	uint32_t l_total = p_left->total_routes;
	uint32_t r_total = p_right->total_routes;

	if (lrank < rrank)
		return -1;

	if (lrank > rrank)
		return 1;

	if (l_total < r_total)
		return -1;

	if (l_total > r_total)
		return 1;

	/* Equal rank, decide by color */
	if (lcolor < rcolor)
		return -1;
	if (lcolor > rcolor)
		return 1;

	/* same color. decide by id */
	if (l_id < r_id)
		return -1;
	return 1;
}

/* when going down, sort according to decreasing order of total routes */
/* we want to serve more loaded switches first */
static int updn_node_load_and_color_cmp_down_func(const void *lhs, const void* rhs)
{
	updn_node_t *p_left = *(updn_node_t **)lhs;
	updn_node_t *p_right = *(updn_node_t **)rhs;
	unsigned lrank = p_left->sw->rank;
	unsigned rrank = p_right->sw->rank;
	color_t lcolor = p_left->color;
	color_t rcolor = p_right->color;
	uint64_t l_id = p_left->sw->id;
	uint64_t r_id = p_right->sw->id;
	uint32_t l_total = p_left->total_routes;
	uint32_t r_total = p_right->total_routes;

	if (lrank < rrank)
		return -1;

	if (lrank > rrank)
		return 1;

	if (l_total > r_total)
		return -1;

	if (l_total < r_total)
		return 1;

	/* Equal rank, decide by color */
	if (lcolor < rcolor)
		return -1;
	if (lcolor > rcolor)
		return 1;

	/* same color. decide by id */
	if (l_id < r_id)
		return -1;
	return 1;
}

static int updn_lid_tracking_build_lft_tables(updn_t *p_updn)
{
	osm_ucast_mgr_t *p_mgr = &p_updn->p_osm->sm.ucast_mgr;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	osm_log_t *p_log = &p_updn->p_osm->log;
	cl_map_item_t *item;
	osm_switch_t *p_sw;
	struct updn_node *u;
	int i, result = 0;
	uint16_t num_lids;
	updn_node_t * prev_adj_sw;
	endport_t *p;
	int port_array_offset, next_offset;
	int prev_rank, max_rank;
	color_t prev_color = MAX_COLOR;
	int found, k;
	int routing_guid_order_used;

	OSM_LOG_ENTER(p_log);

	/* reset new lfts from previous routing */
	osm_ucast_mgr_reset_lfts_all_switches(p_subn);

	if (osm_ucast_mgr_init_port_order_lists(p_mgr, 1)) {
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR AA08: Failed to initialize port order list\n");
		return 1;
	}


	/* Add all ranked updn nodes to an array and Initialize DLID arrays */
	p_updn->num_nodes = 0;
	p_updn->ranked_nodes =
	    (updn_node_t**)calloc(cl_qmap_count(&p_subn->sw_guid_tbl),
				  sizeof(updn_node_t *));
	if (!p_updn->ranked_nodes) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR,
			"ERR AA0D: Failed to allocate memory for ranked nodes"
			" array\n");
		result = 1;
		goto exit;
	}

	for (item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     item = cl_qmap_next(item)) {
		p_sw = (osm_switch_t *)item;
		OSM_ASSERT(p_sw && p_sw->priv);
		u = (updn_node_t *)p_sw->priv;
		if (p_sw->rank == OSM_SW_NO_RANK)
			continue;

		p_updn->ranked_nodes[p_updn->num_nodes++] = u;
		num_lids = u->sw->max_lid_ho + 1;
		u->dlid_routes = (uint32_t *)calloc(num_lids, sizeof(uint32_t));
		u->dlid_sw_routes = (uint32_t *)calloc(num_lids, sizeof(uint32_t));

		if (!u->dlid_routes || !u->dlid_sw_routes) {
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR AA09: Failed to allocates destination LIDs"
				" array\n");
			result = 1;
			goto exit;
		}

		u->total_routes = 0;
	}

	qsort(p_updn->ranked_nodes, p_updn->num_nodes, sizeof(updn_node_t *),
	      updn_node_rank_cmp_func);

	updn_assign_color(p_updn);

	if (create_port_array(p_updn, &p_mgr->port_order_lists[0],
			      p_subn->min_path_bit,
			      p_subn->max_path_bit)) {
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR AA10: Failed to create port array for port list\n");
		result = 1;
		goto exit;
	}

	routing_guid_order_used = p_mgr->guid_routing_order_count;
	/* sort endports by color if no guid_routing_order */
	if(!routing_guid_order_used) {
		qsort(p_updn->port_arr, p_updn->port_arr_len, sizeof(endport_t),
		      updn_port_color_cmp_func);
		/* build offset array */
		p_updn->offset_arr_len = 0;
		prev_adj_sw = NULL;

		for(i=0; i < p_updn->port_arr_len; i++) {
			p = &p_updn->port_arr[i];

			if (p->p_port->p_node->sw) {
				continue;
			}
			if (prev_adj_sw != p->adj_sw) {
				group_offset_t *g =
					&p_updn->offset_arr[p_updn->offset_arr_len];
				prev_adj_sw = p->adj_sw;
				g->offset = i;
				g->color = p->adj_sw->color;
				p_updn->offset_arr_len++;
			}
		}
	}

	next_offset = 0;
	prev_rank = -1;
	prev_color = 0;
	/* From leafs to spines */
	for (i = p_updn->num_nodes - 1; i >= 0; i--) {
		u = p_updn->ranked_nodes[i];
		/*
		* route switches with higher total routes value first
		* under the same rank
		* First sort will be by color as total routes are yet 0
		*/

		if (u->sw->rank != 0 && u->sw->rank != prev_rank) {
			prev_rank = u->sw->rank;
			qsort(&p_updn->ranked_nodes[0], i + 1,
			      sizeof(updn_node_t *),
			      updn_node_load_and_color_cmp_up_func);
			u = p_updn->ranked_nodes[i];
		}

		if (routing_guid_order_used) {
			port_array_offset = 0;
			goto Next1;
		}
		if (u->is_leaf && u->sw->rank) {
			found = 0;
			/* find first offset after the nodes with switch color */
			for (k = 0; k < p_updn->offset_arr_len; k++) {
				while (p_updn->offset_arr[k].color
					== u->color && k < p_updn->offset_arr_len) {
					found = 1;
					k++;
				}
				if (found) {
					next_offset = k % p_updn->offset_arr_len;
					break;
				}
			}
		}
		port_array_offset = p_updn->offset_arr[next_offset].offset;
Next1:
		updn_lid_tracking_process_switch(p_updn, u, UP,
						 port_array_offset);
	}

	/* Clear switch weights */
	for (i = 0; i < p_updn->num_nodes; i++) {
		u = p_updn->ranked_nodes[i];
		if (u->sw->rank != 0 ) {
			num_lids = u->sw->max_lid_ho + 1;
			memset(u->dlid_routes,0, num_lids * sizeof(uint32_t));
			memset(u->dlid_sw_routes,0, num_lids * sizeof(uint32_t));
			u->total_routes = 0;
		}
	}
	prev_rank = -1;
	max_rank = p_updn->ranked_nodes[p_updn->num_nodes - 1]->sw->rank;
	/* From spines to leafs */
	for (i = 0; i < p_updn->num_nodes; i++) {
		u = p_updn->ranked_nodes[i];
		/*
		 * when rank/color changes we need to find first offset
		 * corresponding to the host group with subcolor
		*/
		if (u->sw->rank != max_rank && u->sw->rank != prev_rank) {
			prev_rank = u->sw->rank;
			qsort(&p_updn->ranked_nodes[i], p_updn->num_nodes - i,
			      sizeof(updn_node_t *),
			      updn_node_load_and_color_cmp_down_func);
			u = p_updn->ranked_nodes[i];
		}

		if (routing_guid_order_used) {
			port_array_offset = 0;
			goto Next2;
		}
		if (next_offset == p_updn->offset_arr_len ||
		    prev_rank != u->sw->rank ||
		    prev_color != u->color) {
			prev_rank = u->sw->rank;
			prev_color = u->color;
			for (k = 0; k < p_updn->offset_arr_len; k++) {
				if (p_updn->offset_arr[k].color &
					u->color)
					break;
			}
			next_offset = k;
		}
		/* switch disconnected from topology. Possible with REC */
		if (next_offset == p_updn->offset_arr_len)
			continue;

		port_array_offset = p_updn->offset_arr[next_offset++].offset;
Next2:
		updn_lid_tracking_process_switch(p_updn, u, DOWN,
						 port_array_offset);

	}

	if (!p_subn->opt.updn_lid_tracking_converge_routes)
		goto exit;

	/* Do another pass UP to converge routes  */
	for (i = p_updn->num_nodes - 1; i >= 0; i--) {
		u = p_updn->ranked_nodes[i];
		if (u->sw->rank != 0 )
			updn_lid_tracking_converge_routes(p_updn, u);
	}

exit:
	osm_ucast_mgr_release_port_order_lists(p_mgr, 1);

	OSM_LOG_EXIT(p_log);

	return result;
}

static void updn_calculate_missing_routes(updn_t * p_updn)
{
	osm_ucast_mgr_t *p_mgr = &p_updn->p_osm->sm.ucast_mgr;
	osm_log_t *p_log = &p_updn->p_osm->log;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);

	OSM_LOG_ENTER(p_log);

	if (p_updn->num_roots > 1 &&
	    p_subn->opt.calc_missing_routes &&
	    osm_ucast_mgr_calculate_missing_routes(p_mgr)) {
		OSM_LOG(&p_updn->p_osm->log, OSM_LOG_ERROR,
			"ERR AA06: leaf switches were not set properly, "
			"no missing LID routes will be calculated\n");
	}

	osm_subn_clear_leafs(p_subn);

	OSM_LOG_EXIT(p_log);
}

/* UPDN callback function */
static int updn_fwd_tables(void * ctx)
{
	updn_t *p_updn = ctx;
	osm_subn_t *p_subn = updn_get_subnet(p_updn);
	osm_log_t *p_log = &p_updn->p_osm->log;
	int ret = 0;

	OSM_LOG_ENTER(p_log);

	if (p_subn->opt.updn_lid_tracking_mode) {
		OSM_LOG(p_log, OSM_LOG_VERBOSE,
			"Up/Down routing with LID tracking\n");
		ret = updn_lid_tracking_build_lft_tables(p_updn);
	} else {
		ret = osm_ucast_mgr_build_lft_tables(&p_updn->p_osm->sm.ucast_mgr);
	}

	if (ret < 0)
		goto Exit;

	OSM_LOG(p_log, OSM_LOG_VERBOSE,
		"Connecting switches that are unreachable within "
		"Up/Down rules\n");
	updn_calculate_missing_routes(p_updn);

Exit:
	updn_delete_all(p_updn);

	OSM_LOG_EXIT(p_log);
	return ret;
}

static void updn_delete(void *context)
{
	free(context);
}

static ib_api_status_t updn_mcast_stree(void *context,
					IN OUT struct osm_mgrp_box *mgb)
{
	updn_t *p_updn = (updn_t *)context;
	return osm_mcast_mgr_build_clf_stree(updn_get_subnet(p_updn), mgb);
}

int osm_ucast_updn_setup(struct osm_routing_engine *r, osm_opensm_t *osm)
{
	updn_t *updn;

	updn = malloc(sizeof(updn_t));
	if (!updn)
		return -1;
	memset(updn, 0, sizeof(updn_t));

	updn->p_osm = osm;
	updn->p_subn = osm->sm.ucast_mgr.p_subn;

	r->context = updn;
	r->destroy = updn_delete;
	r->build_lid_matrices = updn_lid_matrices;
	r->ucast_build_fwd_tables = updn_fwd_tables;
	r->mcast_build_stree = updn_mcast_stree;

	return 0;
}

int osm_ucast_ar_updn_setup(struct osm_routing_engine *r,
			    osm_opensm_t *osm)
{
	int rc;

	rc = osm_ucast_updn_setup(r, osm);

	if (rc)
		return rc;

	/* set suitable auto AR parameters */
	r->ar_mode_auto = OSM_ADV_ROUTING_MODE_ENABLED;
	r->shield_mode_auto = OSM_ADV_ROUTING_MODE_RN;
	r->pfrn_mode = TRUE;
	r->adv_routing_engine_auto = OSM_ADV_ROUTING_TYPE_AR_TREE;

	return 0;
}
